// Utility to register/unregister content of cache to Indexing Service (RC,RLS)
#include "../std.h"
#include <globus_common.h>

#include "../cache/cache.h"
#include "../config/config_file.h"
#include "../config/environment.h"
#include "../jobs/users.h"
#include "../datamove/datapoint.h"
#include "../misc/log_time.h"


int main(int argc,char* argv[]) {
  LogTime::active=true;
  LogTime::level=0;
  for(;;) {
    opterr=0;
    int optc=getopt(argc,argv,"+hd:c:");
    if(optc == -1) break;
    switch(optc) {
      case 'h': {
        olog<<"cache-register [-h] [-d level] [-c configuration_file]"<<endl;
        olog<<" -h - print help"<<endl;
        olog<<" -d - set debug level (use 0 for verbose)"<<endl;
        olog<<" -c - path to grid-manager's configuration file"<<endl;
        exit(1);
      }; break;
      case 'v': {
        olog<<"cache-register: version "<<VERSION<<endl;
        exit(0);
      };
      case 'd': {
        char* p;
        int i = strtol(optarg,&p,10);
        if(((*p) != 0) || (i<0)) {
          olog<<"Improper debug level '"<<optarg<<"'"<<endl;
          exit(1);
        };
        LogTime::level=i;
      }; break;
      case 'c': {
        nordugrid_config_loc=optarg;
      }; break;
      case '?': {
        olog<<"Unsupported option '"<<(char)optopt<<"'"<<endl;
        exit(1);
      }; break;
      case ':': {
        olog<<"Missing parameter for option '"<<(char)optopt<<"'"<<endl;
        exit(1);
      }; break;
      default: {
        olog<<"Undefined processing error"<<endl;
        exit(1);
      };
    };
  };

  char hostname[256];
  if(globus_libc_gethostname(hostname,sizeof(hostname)) != 0) {
    cerr<<"Failed to obtain own host name"<<endl;
    return -1;
  };
  string cache_name(hostname);
  string cache_url(hostname);
  cache_name=cache_name+":cache";
  cache_url="cache://"+cache_url+"/";
  // Process configuration (like in GM)
  if(!read_env_vars()) exit(1);
  JobUsers users;
  string my_username("");
  uid_t my_uid=getuid();
  JobUser *my_user = NULL;
  {
    struct passwd pw_;
    struct passwd *pw;
    char buf[BUFSIZ];
    getpwuid_r(my_uid,&pw_,buf,BUFSIZ,&pw);
    if(pw != NULL) { my_username=pw->pw_name; };
  };
  if(my_username.length() == 0) {
    olog << "Can't recognize myself." << endl; exit(1);
  };
  my_user = new JobUser(my_username);
  if(!configure_serviced_users(users,my_uid,my_username,*my_user)) {
    cout<<"Error processing configuration."<<endl; exit(1);
  };
  if(users.size() == 0) {
    cout<<"No suitable users found in configuration."<<endl; exit(1);
  };

  // Process all configuread caches
  for(JobUsers::const_iterator user = users.begin();user != users.end();++user) {
    if(user->CachePrivate()) continue; // only public caches made public
    string cache_dir = user->CacheDir();
    if(cache_dir.length() == 0) continue; // cache not configured
    olog<<"Cache directory: "<<cache_dir<<endl;

    // Get list of old and new files (snapshot)
    list<string> old_records;
    list<string> new_records;
    if(cache_history_lists(cache_dir.c_str(),old_records,new_records) != 0) {
      cerr<<"Failed to get lists from cache"<<endl;
      return -1;
    };

    // Remove duplicates


    // "old" records have priority


    // Process "new"
    for(list<string>::iterator i = new_records.begin();i!=new_records.end();) {
      cout<<"Record (new): "<<*i<<endl;
      DataPoint url(i->c_str()); // take url
      if(url.meta()) { // ordinary urls are not registrable
        // Add new location. This also limits resolved locations 
        // to those specified
        if(!url.add_location(cache_name.c_str(),(cache_url+url.lfn()).c_str())) {
          cerr<<"Failed to add location: "<<url<<endl; 
          i=new_records.erase(i); continue;
        };
        //  Resolve for destination to check if name is registered
        //  and for metadata (would be nice to get already available data)
        if(!url.meta_resolve(false)) {
          cerr<<"Failed to resolve: "<<url<<endl;
          i=new_records.erase(i); continue;
        };
        // preregister for replication
        if(!url.meta_preregister(true)) {
          cerr<<"Failed to preregister: "<<url<<endl;
          i=new_records.erase(i); continue;
        };
        // postregister to announce existence of file
        if(!url.meta_postregister(true,false)) {
          cerr<<"Failed to postregister: "<<url<<endl;
          i=new_records.erase(i); continue;
        };
      };
      ++i;
    };  

    // Process "old"
    for(list<string>::iterator i = old_records.begin();i!=old_records.end();) {
      cout<<"Record (old): "<<*i<<endl;
      DataPoint url(i->c_str()); // take url
      if(url.meta()) { // ordinary urls are not registrable
        // Add new location. This also limits resolved locations 
        // to those specified
        if(!url.add_location(cache_name.c_str(),(cache_url+url.lfn()).c_str()))
{
          cerr<<"Failed to add location: "<<url<<endl;
          i=old_records.erase(i); continue;
        };
        //  Resolve for source to check if name is registered
        if(!url.meta_resolve(true)) {
          cerr<<"Failed to resolve: "<<url<<endl;
          i=old_records.erase(i); continue;
        };
        if(!url.have_locations()) {
          cout<<"Location is already missing (that's ok)"<<endl;
        } else {
          if(!url.meta_unregister(false)) {
            odlog(0)<<"Failed to unregister"<<endl;
            i=old_records.erase(i); continue;
          };
        };
      };
      ++i;
    };

    // Remove processed records
    if(cache_history_remove(cache_dir.c_str(),old_records,new_records) != 0) {
      cerr<<"Failed to remove lists from cache"<<endl;
      return -1;
    };
  };
  return 0;
}


