#include "../std.h"
#include "../config/config_map.h"
#include "../transfer/replica_utils.h"
#include "../transfer/replica.h"
#include "../misc/url_options.h"
#include "../misc/checksum.h"
#include "../misc/log_time.h"

#include "datapoint.h"

/* perform resolve operation, which can take long time */
bool DataPoint::meta_resolve_rc(bool source) {
    is_resolved=false;
    /* connect to RC server and replace locations with urls */
    is_metaexisting=false;
    if(!rc_mgr) rc_mgr = new RCManager(rc_url,"","");
    if(!(rc_mgr->is_open())) {
      odlog(0)<<"Failed accessing Replica Catalog collection: "<<rc_url<<endl;
      delete rc_mgr; rc_mgr=NULL;
      return false;
    };
    list<RCLocation> *locs = new list<RCLocation>;
   /* !!!!! TODO - make sure results returned by rc_mgr are real results
      and not errors */ 
    odlog(2)<<"meta_resolve: rc_lfn: "<<rc_lfn<<endl;
    if(source) {
      odlog(2)<<"meta_resolve: for source"<<endl;
      if(!rc_mgr->GetLocations("",*locs,rc_lfn.c_str())) { 
        odlog(0)<<"Locations not found: "<<rc_url<<endl;
        delete locs; delete rc_mgr; rc_mgr=NULL; return false;
      };
      is_metaexisting=true;
    }
    else {
      odlog(2)<<"meta_resolve: for destination"<<endl;
      if(!(rc_mgr->GetLocations("",*locs))) {
        odlog(0)<<"Locations not found: "<<rc_url<<endl;
        delete locs; delete rc_mgr; rc_mgr=NULL; return false;
      };
    };
    if(locations.size() == 0) { // no predefined locations - accept all found
      list<RCLocation>::iterator rcloc;
      for(rcloc=locs->begin();rcloc!=locs->end();++rcloc) {
        locations.push_back(Location(rcloc->name,rcloc->url+rc_lfn));
      };
      for(list<Location>::iterator loc=locations.begin();loc!=locations.end();++loc) {
          if(common_url_options.length() != 0)
            add_url_options(loc->url,common_url_options.c_str(),-1);
      };
    }
    else {
      list<Location>::iterator loc = locations.begin();
      list<RCLocation>::iterator rcloc;
      for(;loc!=locations.end();) {
        bool found = false;
        string name = loc->meta;
        string options = "";
        int n = name.find(';'); // remove options from name for clean compare
        if(n != string::npos) {
          options=name.substr(n+1);
          name.resize(n);
        };
        for(rcloc=locs->begin();rcloc!=locs->end();++rcloc) {
          if(rcloc->name != name) continue;
          loc->url=rcloc->url+rc_lfn;
          loc->meta=name; // effectively strips options off :)
          loc->existing=true; // if it was addded manualy, now it is resolved
          locs->erase(rcloc);
          if(common_url_options.length() != 0)
            add_url_options(loc->url,common_url_options.c_str(),-1);
          add_url_options(loc->url,options.c_str(),-1);
          found=true;
          break;
        };
        if(!found) { 
          if(loc->existing) { loc=locations.erase(loc); } // remove resolved
          else { ++loc; }; // and keep manual
        }
        else { ++loc; };
      };
    };
    /* obtain metadata */
    RCFile file(rc_lfn,0,false,0,false,0,false);
    if(!rc_mgr->GetFile(file)) {
      if(source) {
        odlog(2)<<"File "<<rc_lfn<<" does not exist or has no metadata"<<endl;
      };
    }
    else {
      is_metaexisting=true;
      odlog(2)<<"meta_get_data: obtained metadata"<<endl;
      if(!meta_checksum_valid) meta_checksum(file.get_checksum());
      if(!meta_size_valid) meta_size(file.get_size());
      if(!meta_created_valid) meta_created(file.get_timestamp());
    };
    odlog(2)<<"meta_get_data: checksum: "<<meta_checksum()<<endl;
    odlog(2)<<"meta_get_data: size: "<<meta_size()<<endl;
    odlog(2)<<"meta_get_data: created: "<<meta_created()<<endl;
    delete locs; delete rc_mgr; rc_mgr=NULL;
    location=locations.begin();
    is_resolved=true;
    return true;
}

bool DataPoint::meta_preregister_rc(bool replication,bool force) {
    odlog(2)<<"meta_preregister: is rc"<<endl;
    if(replication) { /* replicating inside same lfn */
      if(!is_metaexisting) { /* for replication it must be there */
        odlog(0)<<"LFN is missing in replica (needed for replication)"<<endl;
        return false;
      };
      return true;
    };
    if(is_metaexisting) { /* algorithm require this to be new file */
      if(!force) { /* unless we have dead SE */
        odlog(0)<<"LFN already exists in replica (should not for new file)"<<endl;
        return false;
      };
    };
    if(!meta_size_valid) {
      odlog(0)<<"Not enough metadata available for lfn creation"<<endl;
      return false;
    };
    rc_mgr = new RCManager(rc_url,"","");
    if(!(rc_mgr->is_open())) {
      odlog(0)<<"Failed accessing Replica Catalog collection: "<<rc_url<<endl;
      delete rc_mgr; rc_mgr=NULL; return false;
    };
    odlog(2)<<"meta_preregister: creating RCFile"<<endl;
    RCFile file(rc_lfn,
           meta_size(),meta_size_valid,
           meta_checksum(),meta_checksum_valid,
           meta_created(),meta_created_valid);
/* !!! TODO - make sure results returned by rc_mgr are real results
      and not errors */
    odlog(2)<<"meta_preregister: creating lfn in RC"<<endl;
    if(!rc_mgr->AddFile(file,!force)) {
      if(!force) {
        odlog(0)<<"Failed to create lfn in RC"<<endl;
        delete rc_mgr; rc_mgr=NULL; return false;
      };
      // if force, then just add missing metadata
      if(!rc_mgr->UpdateFile(file)) {
        odlog(0)<<"Failed to update lfn in RC"<<endl;
        delete rc_mgr; rc_mgr=NULL; return false;
      };
    };
    delete rc_mgr; rc_mgr=NULL;
    is_metaexisting=true;
    return true;
}

bool DataPoint::meta_postregister_rc(bool replication,bool failure) {
    if(!is_metaexisting) { 
      odlog(0)<<"File was not yet registered in replica"<<endl;
      return false;
    };
    rc_mgr = new RCManager(rc_url,"","");
    if(!(rc_mgr->is_open())) {
      odlog(0)<<"Failed accessing Replica Catalog collection: "<<rc_url<<endl;
      delete rc_mgr; rc_mgr=NULL; return false;
    };
    odlog(2)<<"meta_postregister: updating RCFile"<<endl;
    RCFile file(rc_lfn,
           meta_size(),meta_size_valid,
           meta_checksum(),meta_checksum_valid,
           meta_created(),meta_created_valid);
/* !!! TODO - make sure results returned by rc_mgr are real results
      and not errors */
    if(!rc_mgr->UpdateFile(file)) {
      odlog(0)<<"Failed to update lfn attributes in RC"<<endl;
      rc_mgr->RemoveFile(file);
      delete rc_mgr; rc_mgr=NULL; return false;
    };
    string host = location->meta;
    odlog(2)<<"meta_postregister: creating pfn in RC"<<endl;
    string loc_url("*");
    if(!(location->existing)) {
      // manually added location possibly need registration
      loc_url=location->url;
      loc_url.resize(location->url.length()-rc_lfn.length());
    };
    if(!rc_mgr->AddFileLocation(file,host,loc_url.c_str(),true)) {
      odlog(0)<<"Failed to create pfn in RC"<<endl;
      // delete 
      if(!replication) rc_mgr->RemoveFile(file);
      delete rc_mgr; rc_mgr=NULL; return false;
    };
    delete rc_mgr; rc_mgr=NULL;
    return true;
}

bool DataPoint::meta_preunregister_rc(bool replication) {
  if(replication) return true;
  if(meta_unregister_rc(true)) {
    is_metaexisting=false;
    return true;
  };
  return false;
}

bool DataPoint::meta_unregister_rc(bool all) {
/* not fully implemented yet - all=true will remove logical file only */
    if(!all) {
      if(location == locations.end()) { 
        odlog(0)<<"Location is missing"<<endl;
        return false;
      };
    };
    rc_mgr = new RCManager(rc_url,"","");
    if(!(rc_mgr->is_open())) {
      odlog(0)<<"Failed accessing Replica Catalog collection: "<<rc_url<<endl;
      delete rc_mgr; rc_mgr=NULL; return false;
    };
    odlog(2)<<"meta_unregister: creating RCFile"<<endl;
    RCFile file(rc_lfn,0,false,0,false,0,false);
    if(all) {
      if(rc_mgr->RemoveFile(file) != true) {
        odlog(0)<<"Failed to remove lfn in RC"<<endl;
        delete rc_mgr; rc_mgr=NULL; return false;
      };
    }
    else {
      if(rc_mgr->RemoveFileLocation(file,location->meta) != true) {
        odlog(0)<<"Failed to remove pfn in RC"<<endl;
        delete rc_mgr; rc_mgr=NULL; return false;
      };
    };
    delete rc_mgr; rc_mgr=NULL;
    return true;
}

/* extract info from metacatalog url */
bool DataPoint::process_meta_url_rc(const string &url,string &cat_url,string &filename,list<DataPoint::Location> &locations) {
  if(!strncasecmp(url.c_str(),"rc://",5)) {
    cat_url.resize(0); locations.clear(); filename.resize(0);
    string url_(url.c_str());
    if(!extract_RC_from_url(url_,cat_url)) {
      odlog(0)<<"URL part to contact Replica Catalog is missing in "<<url<<endl;
      return false;
    };
    /* extract host and filename from url */
    const char* s=url_.c_str()+strlen(rc_url_head);
    const char* s_e = strchr(s,'/'); if(s_e == NULL) s_e=s+strlen(s);
    string host(s,s_e-s);
    if(host.length()) if(host[host.length()-1] == '@') host.resize(host.length()-1);
    odlog(2)<<"Location hosts: "<<host<<endl;
    int n=0;
    for(int nn=0;n<host.length();) {
      nn=host.find('|',n); if(nn == string::npos) nn=host.length();
      if(n == nn) { n++; continue; };
      if(host.c_str()[n] == ';') {
        common_url_options+=host.substr(n,nn-n);
      }
      else {
        string loc(host.c_str()+n,nn-n);
        locations.push_back(DataPoint::Location(loc.c_str(),""));
      };
      n=nn+1;
    };
    if(common_url_options.length() != 0)
      if(common_url_options[0] == ';') 
        common_url_options.erase(0,1);
    const char* fname = NULL;
    if((*s_e) != 0) {
      filename=(s_e+1);
      odlog(2)<<"Logical filename: "<<filename<<endl;
    };
    return true;
  };
  return false;
};

// obtain list of lfns and corresponding attributes
bool DataPoint::list_files_rc(list<DataPoint::FileInfo> &files,bool resolve) {
    if(!rc_mgr) rc_mgr = new RCManager(rc_url,"","");
    if(!(rc_mgr->is_open())) {
      odlog(0)<<"Failed accessing Replica Catalog collection: "<<rc_url<<endl;
      delete rc_mgr; rc_mgr=NULL;
      return false;
    };
    list<RCFile> rcfiles;
   /* !!!!! TODO - make sure results returned by rc_mgr are real results
      and not errors */
odlog(2)<<"list_files_rc: retrieving lfns"<<endl;
    if(!rc_mgr->ListFiles(rcfiles)) {
      odlog(0)<<"Failed to list files: "<<rc_url<<endl;
      delete rc_mgr; rc_mgr=NULL; return false;
    };
    string tmp_rc_lfn = rc_lfn;
    bool result = true;
    for(list<RCFile>::iterator rcf=rcfiles.begin();rcf!=rcfiles.end();++rcf) {
      list<DataPoint::FileInfo>::iterator f = 
             files.insert(files.end(),DataPoint::FileInfo(rcf->get_name()));
odlog(2)<<"list_files_rc: lfn: "<<f->name<<endl;
      if(resolve) {
        is_resolved=false;  // hack !!!
        meta_size_valid=false; meta_checksum_valid=false;
        meta_created_valid=false; meta_validtill_valid=false;
        rc_lfn=f->name;
        if(!get_info(*f)) result=false;
        is_resolved=false;
        meta_size_valid=false; meta_checksum_valid=false;
        meta_created_valid=false; meta_validtill_valid=false;
      };
    };
    rc_lfn=tmp_rc_lfn;
    return result;
}

