#include "../std.h"
#include "fileroot.h"
#include "conf.h"
#include "names.h"
#include "misc.h"
#include "../misc/canonical_dir.h"
#include "../misc/log_time.h"

int FileRoot::config(ifstream &cfile,string &pluginpath) {
  bool right_group = true;
  for(;;) {
    string rest;
    string command=config_read_line(cfile,rest);
    if(command.length() == 0) break; /* EOF */
    if(command == "include") {  /* include content of another file */
      string name=config_next_arg(rest);
      ifstream cfile_;
      cfile_.open(name.c_str(),ifstream::in);
      if(!cfile_.is_open()) {
        olog << "Error: couldn't open file " << name << endl;
        return 1;
      };
      config(cfile_,pluginpath);
      cfile_.close();
    }
    else if(command == "encryption") {  /* is encryption allowed ? */
      string value=config_next_arg(rest);
      if(value == "yes") {
        heavy_encryption=true;
      } else if(value == "no") {
        heavy_encryption=false;
      } else {
        user.user.clear_groups();
        nodes.clear();
        olog<<"Error: improper attribute for encryption command: "<<value<<endl;
        return 1;
      };
    }
    else if(command == "allowunknown") { /* should user be present in grid-mapfile ? */
      string value=config_next_arg(rest);
      if(value == "yes") {
        user.gridmap=true;
      } else if(value == "no") {
        if(!user.gridmap) {
          olog << "Error: unknown (non-gridmap) user is not allowed" << endl;
          return 1;
        };
      } else {
        user.user.clear_groups();
        nodes.clear();
        olog<<"Error: improper attribute for encryption command: "<<value<<endl;        return 1;
      };
    }
    else if(command == "group") { /* definition of group of users based on DN */
      if(!right_group) {
        for(;;) {
          command=config_read_line(cfile,rest);
          if(command.length() == 0) break; /* eof - bad */
          if(command == "end") break;
        };
        continue;
      };
      string group_name=config_next_arg(rest);
      int decision = AAA_NO_MATCH;
      for(;;) {
        string rest=config_read_line(cfile);
        if(rest.length() == 0) break; /* eof - bad */
        if(rest == "end") break;
        if(decision == AAA_NO_MATCH) {
          decision = user.user.evaluate(rest.c_str());
          if(decision == AAA_FAILURE) decision=AAA_NO_MATCH;
        };
      };
      if(decision == AAA_POSITIVE_MATCH) user.user.add_group(group_name);
    }
    else if(command == "unixmap") {  /* map to local unix user */
      if(!unix_mapped) {
        string unix_name=config_next_arg(rest);
        string unix_group="";
        int n = unix_name.find(':');
        if(n != string::npos) {
          unix_group=unix_name.substr(n+1,string::npos);
          unix_name.erase(n,string::npos);
        };
        int decision = user.user.evaluate(rest.c_str());
        if(decision == AAA_POSITIVE_MATCH) {
          if(!user.unix_user(unix_name.c_str(),unix_group.c_str())) {
            olog<<"Wrong user specified: "<<unix_name<<" "<<unix_group<<endl;
            return 1;
          };
          unix_mapped=true;
        };
      };
    }
    else if(command == "groupcfg") {  /* next commands only for these groups */
      if(rest.find_first_not_of(" \t") == string::npos) {
        right_group=true; continue;
      };
      right_group=false;
      for(;;) {
        string group_name=config_next_arg(rest);
        if(group_name.length() == 0) break;
        right_group=user.user.check_group(group_name);
        if(right_group) break;
      };
    }
    else if(command == "pluginpath") {
      if(!right_group) continue;
      pluginpath=config_next_arg(rest);
      if(pluginpath.length() == 0) pluginpath="/";
    }
    else if(command == "plugin") {
      if(!right_group) {
        for(;;) {
          command=config_read_line(cfile,rest);
          if(command.length() == 0) break; /* eof - bad */
          if(command == "end") break;
        };
        continue;
      };
      string dir = config_next_arg(rest);
      string plugin = config_next_arg(rest);
      if(plugin.length() == 0) {
        olog << "Warning: can't parse configuration line" << endl;
        continue;
      };
      dir=subst_user_spec(dir,&user);
      if(canonical_dir(dir,false) != 0) {
        olog << "Warning: bad directory in plugin directive" << endl;
        continue;
      };
      /* look if path is not alredy registered */
      bool already_have_this_path=false;
      for(list<FileNode>::iterator i=nodes.begin();i!=nodes.end();++i) {
        if((*i) == dir) {
          already_have_this_path=true;
          break;
        };
      };
      if(already_have_this_path) {
        olog << "Already have directory:"<<dir<<endl;
        for(;;) {
          command=config_read_line(cfile,rest);
          if(command.length() == 0) break; /* eof - bad */
          if(command == "end") break;
        };
      }
      else {
        olog<<"Registering directory:"<<dir<<" with plugin: "<<plugin<<endl;
        plugin=pluginpath+'/'+plugin;
        FileNode node((char*)(dir.c_str()),(char*)(plugin.c_str)(),cfile,user);
        if(node.has_plugin()) { nodes.push_back(node); }
        else {
          olog << "Error: file node creation failed: " << dir << endl;
          for(;;) {
            command=config_read_line(cfile,rest);
            if(command.length() == 0) break; /* eof - bad */
            if(command == "end") break;
          };
        };
      };
    }
    else {
      olog << "Warning: unsupported configuration command" << endl;
    };
  };
  return 0;
}

int FileRoot::config(globus_ftp_control_auth_info_t *auth,
                     gss_cred_id_t delegated_cred,
                     globus_ftp_control_handle_t *handle) {
  /* open and read configuration file */
  ifstream cfile;
  config_open(cfile);
  if(!cfile.is_open()) {
    olog << "Error: configuration file not found" << endl;
    return 1;
  };
  /* keep information about user */
  if(fill_user_spec(&user,auth,delegated_cred,handle) != 0) return 1;
  string pluginpath="/";
  int r = config(cfile,pluginpath);
  config_close(cfile);
  if(r != 0) return r;
  if(!user.gridmap) {
    olog << "Error: unknown (non-gridmap) user is not allowed" << endl;
    return 1;
  };
  /* must be sorted to make sure we can always find right directory */
  nodes.sort(FileNode::compare);
  /* create dummy directories */
  int nn=nodes.size();
  for(list<FileNode>::iterator i=nodes.begin();i!=nodes.end();++i) {
    if(nn==0) break;
    string name = i->point;
    for(;remove_last_name(name);) {
      if(name.length() == 0) break;
      bool found = false;
      list<FileNode>::iterator ii=i;
      for(;;) {
        ++ii; if(ii == nodes.end()) break;
        if(name == ii->point) { found=true; break; };
      };
      if(!found) { /* add dummy dir */
        olog<<"Registering dummy directory:"<<name<<endl;
        nodes.push_back(FileNode(name.c_str()));
      }
      else {
        break;
      };
    };
    nn--;
  };
  opened_node=nodes.end();
  user.free();
  return 0;
}

