#include "../../std.h"

#ifdef HAVE_SSTREAM
#include <sstream>
#else
#include <strstream>
#endif

#define GRIDFTP_PLUGIN
#include "../../config/conf.h"
#include "../../jobs/job.h"
#include "../../jobs/plugins.h"
#include "../userspec.h"
#include "../names.h"
#include "../misc.h"
#include "../../config/environment.h"
#include "../../config/config_pre.h"
#include "../../files/info_types.h"
#include "../../rsl/parse_rsl.h"
#include "../../misc/inttostring.h"
#include "../../misc/delete.h"
#include "../../misc/escaped.h"
#include "../../misc/log_time.h"
#include "../../transfer/proxy.h"
#include "../../run/run.h"
#include "../fileplugin/fileplugin.h"
#include "CertInfo.h"
#include "jobplugin.h"

RunParallel JobPlugin::run;

JobPlugin::JobPlugin(istream &cfile,userspec_t &user_s) {
//  olog << "JobPlugin: constructor" << endl;
  initialized=true;
  rsl_opened=false;
  user=NULL;
  direct_fs=NULL;
  proxy_fname="";
  string configfile("");
  cont_plugins=NULL;
  readonly=false;
  for(;;) {
    string rest;
    string command=config_read_line(cfile,rest);
    if(command.length() == 0) { break; } /* end of file - should not be here */
    else if(command == "configfile") {
      input_escaped_string(rest.c_str(),configfile); 
    } else if(command == "allownew") {
      string value("");
      input_escaped_string(rest.c_str(),value);
      if(strcasecmp(value.c_str(),"no") == 0) { readonly=true; }
      else if(strcasecmp(value.c_str(),"yes") == 0) { readonly=false; };
    } else if(command == "end") {
      break; /* end of section */
    }
    else {
      olog<<"Warning: unsupported configuration command: "<<command<<endl;
    };
  };
  if(!run.is_initialized()) {
    olog<<"Warning: Initialization of signal environment failed"<<endl;
  };
  if(!read_env_vars()) { initialized=false; }
  {
    if(configfile.length()) nordugrid_config_loc=configfile;
    user=new JobUser(user_s.get_uname());
    if(!user->is_valid()) { initialized=false; }
    else {
      /* read configuration */
      if(nordugrid_loc.length() != 0) {
        string session_root;
        string control_dir;
        string default_lrms;
        string default_queue;
        cont_plugins = new ContinuationPlugins;
        if(!configure_user_dirs(user_s.get_uname(),
          control_dir,session_root,
          default_lrms,default_queue,*cont_plugins)) {
          olog<<"Failed processing grid-manager configuration"<<endl;
          initialized=false;
        }
        else {
          user->SetControlDir(control_dir);
          user->SetSessionRoot(session_root);
          user->SetLRMS(default_lrms,default_queue);
        };
        /* link to the class for direct file access */
        string direct_config = "mount "+session_root+"\n";
        direct_config+="dir / nouser read cd dirlist delete append overwrite";
        direct_config+=" create "+
           inttostring(user->get_uid())+":"+inttostring(user->get_gid())+
           " 600:600";
        direct_config+=" mkdir "+
           inttostring(user->get_uid())+":"+inttostring(user->get_gid())+
           " 700:700\n";
        direct_config+="end\n";
#ifdef HAVE_SSTREAM
        stringstream fake_cfile(direct_config);
#else
        strstream fake_cfile;
        fake_cfile<<direct_config;
#endif
        direct_fs = new DirectFilePlugin(fake_cfile,user_s);
      }
      else {
        initialized=false;
      };
    };
  };
  if(!initialized) if(user) { delete user; user=NULL; };
  if((user_s.user.proxy() == NULL) || (user_s.user.proxy()[0] == 0)) {
    olog<<"WARNING: no delegated credentials were passed"<<endl;
  } else {
    proxy_fname=user_s.user.proxy();
  };
  subject=user_s.user.DN();
  port=user_s.get_port();
  memcpy(host,user_s.get_host(),sizeof(host));
  job_id="";
}

JobPlugin::~JobPlugin(void) {
  delete_job_id();
  if(proxy_fname.length() != 0) { remove(proxy_fname.c_str()); };
  if(cont_plugins) delete cont_plugins;
//  olog << "JobPlugin: destructor " << count << endl;
};
 

int JobPlugin::makedir(string &dname) {
  if(!initialized) return 1;
  if(is_allowed(dname.c_str(),true)) {
    return direct_fs->makedir(dname);
  };
  return 1;
}

int JobPlugin::removefile(string &name) {
  if(!initialized) return 1;
  if(name.find('/') == string::npos) { /* request to cancel the job */
    if(is_allowed(name.c_str())) {  /* owner of the job */
      JobId id(name); JobDescription job_desc(id,"");
      if(job_cancel_mark_put(job_desc,*user)) return 0;
    };
    return 1;
  };
  const char* logname;
  if(is_allowed(name.c_str(),false,NULL,&logname)) {
    if(logname) {
      if((*logname) != 0) return 0; /* pretend status file is deleted */
    };
    return direct_fs->removefile(name);
  };
  return 1;
}

int JobPlugin::removedir(string &dname) {
  if(!initialized) return 1;
  if(dname.find('/') == string::npos) { /* request to clean the job */
    if(is_allowed(dname.c_str())) {  /* owner of the job */
      /* check the status */
      JobId id(dname); 
      job_state_t status=job_state_read_file(id,*user);
      if(status == JOB_STATE_FINISHED) { /* remove files */
        if(job_clean_final(id,*user)) return 0;
      }
      else { /* put marks */
        JobDescription job_desc(id,"");
        bool res = job_cancel_mark_put(job_desc,*user);
        res &= job_clean_mark_put(job_desc,*user);
        if(res) return 0;
      };
    };
    return 1;
  };
  if(is_allowed(dname.c_str())) {
    return direct_fs->removedir(dname);
  };
  return 1;
}

int JobPlugin::open(const char* name,open_modes mode,unsigned long long int size) {
  olog << "jobplugin: open: " << name << endl;
  if(!initialized) return 1;
  if(rsl_opened) {  /* unfinished request - cancel */
    olog<<"Request to open file with storing in progress"<<endl;
    rsl_opened=false;
    delete_job_id();
    return 1; 
  };
  if( mode == GRIDFTP_OPEN_RETRIEVE ) {  /* open for reading */
    string fname;
    const char* logname;
    /* check if reading status files */
    if(is_allowed(name,false,&fname,&logname)) {
      if(logname) {
        if((*logname) != 0) {
          fname=user->ControlDir()+"/job."+fname+"."+logname;
          return direct_fs->open_direct(fname.c_str(),mode);
        };
      };
      return direct_fs->open(name,mode);
    };
    return 1;
  }
  else if( mode == GRIDFTP_OPEN_STORE ) {
    string name_f(name);
    int n = name_f.find('/');
    if((n != string::npos) && (n != 0)) {
      if((strncmp(name_f.c_str(),"new",n) == 0) ||
         (strncmp(name_f.c_str(),job_id.c_str(),n) == 0)) {
        if(name_f.find('/',n+1) != string::npos) return 1; /* should not contain subdirs */
        if(job_id.length() == 0) {
          if(!make_job_id()) {
            olog<<"Failed to allocate id for job"<<endl;
            return 1;
          };
        };
        olog<<"Accepting submission of new job: "<<job_id<<endl;
        memset(job_rsl,0,sizeof(job_rsl));
        rsl_opened=true;
        return 0;
      };
    };
/*
    int n = strlen("new/");  * new jobs should be uploaded to /new *
    if(!strncmp(name,"new/",n)) {
      if(strchr(name+n,'/') != NULL) return 1; * should not contain subdirs *
      if(job_id.length() == 0) { if(!make_job_id()) return 1; };
      memset(job_rsl,0,sizeof(job_rsl));
      rsl_opened=true;
      return 0;
    };
*/
    if(is_allowed(name,true)) {
      return direct_fs->open(name,mode,size);
    };
    return 1;
  }
  else {
    olog << "Warning: unknown open mode " << mode << endl;
    return 1;
  };
}

int JobPlugin::close(bool eof) {
  if(!initialized) return 1;
  if(!rsl_opened) { return direct_fs->close(eof); };
  /* store RSL */
  rsl_opened=false;
  if(job_id.length() == 0) return 1;
  if(!eof) { job_id=""; return 0; }; /* download was canceled */
  string rsl_fname=user->ControlDir()+"/job."+job_id+".description";
  {
    int l;
    int h=::open(rsl_fname.c_str(),O_WRONLY | O_CREAT | O_EXCL,0600);
    if(h == -1) { l=-1; }
    else {
      const char* s;
      for(l=0,s=job_rsl;(*s) && (l!=-1);s+=l) l=::write(h,s,strlen(s));
      ::close(h);
    };
    if(l == -1) { 
      olog<<"Failed writing RSL"<<endl;
      remove(rsl_fname.c_str());
      delete_job_id(); return 1;
    };
  };
  /* analyze rsl (checking, substituting, etc)*/
/*
  string action;
  string jobid_;
  string lrms;
  string queue;
  if(!parse_rsl_for_action(rsl_fname.c_str(),action,jobid_,lrms,queue)) { 
*/
  JobLocalDescription job_desc;
  if(!parse_rsl(rsl_fname.c_str(),job_desc)) {
    olog<<"Failed parsing RSL"<<endl;
    remove(rsl_fname.c_str());
    delete_job_id();
    return 1;
  };
  if(job_desc.action != "request") {
    olog<<"action("<<job_desc.action<<") != request"<<endl;
    remove(rsl_fname.c_str());
    delete_job_id();
    return 1;
  };
  if(job_desc.queue.length() == 0) job_desc.queue=user->DefaultQueue();
  if(job_desc.lrms.length() == 0) job_desc.lrms=user->DefaultLRMS();
  if(job_desc.jobid.length() != 0) {
    delete_job_id(); 
    if(!make_job_id(job_desc.jobid)) {
      olog<<"Failed allocating requested job id: "<<job_desc.jobid<<endl;
      remove(rsl_fname.c_str());
      return 1;
    };
    string fname=user->ControlDir()+"/job."+job_id+".description";
    if(rename(rsl_fname.c_str(),fname.c_str()) != 0) {
      olog<<"Failed moving RSL"<<endl;
      remove(rsl_fname.c_str());
      delete_job_id(); 
      return 1;
    };
    rsl_fname=fname;
  };
  if(!preprocess_rsl(rsl_fname,user->SessionRoot()+"/"+job_id,job_id)) {
    olog<<"Failed preprocessing RSL"<<endl;
    remove(rsl_fname.c_str());
    delete_job_id(); 
    return 1;
  };
  /* create local file */
  /* !!!!! some parameters are uncheked here - rerun,diskspace !!!!! */
/*
  job_desc.lrms=lrms;
  job_desc.queue=queue;
*/
  job_desc.jobid=job_id;
  job_desc.starttime=time(NULL);
  job_desc.DN=subject;
  if(port != 0) {
    job_desc.clientname=
       inttostring(host[0])+"."+inttostring(host[1])+"."+
       inttostring(host[2])+"."+inttostring(host[3])+":"+
       inttostring(port);
  };
  JobDescription job(job_id,"",JOB_STATE_ACCEPTED);
  if(!job_local_write_file(job,*user,job_desc)) {
    olog<<"Failed writing local description"<<endl;
    remove(rsl_fname.c_str());
    delete_job_id(); 
    return 1;
  };
  /* try to create proxy */
  if(proxy_fname.length() != 0) {
    string fname=user->ControlDir()+"/job."+job_id+".proxy";
    int h=::open(fname.c_str(),O_WRONLY | O_CREAT | O_EXCL,0600);
    if(h == -1) return 1;
    int hh=::open(proxy_fname.c_str(),O_RDONLY);
    if(hh == -1) {
      ::close(h);
      ::remove(fname.c_str());
      return 1;
    };
    fix_file_owner(fname,*user);
    int l,ll;
    const char* s;
    char buf[256]; 
    for(;;) {
      ll=::read(hh,buf,sizeof(buf));
      if((ll==0) || (ll==-1)) break;
      for(l=0,s=buf;(ll>0) && (l!=-1);s+=l,ll-=l) l=::write(h,s,ll);
      if(l==-1) break;
    };
    ::close(h);
    ::close(hh);
  };
  /* create status file */
  if(!job_state_write_file(job,*user,JOB_STATE_ACCEPTED)) {
    olog<<"Failed writing status"<<endl;
    remove(rsl_fname.c_str());
    string fname=user->ControlDir()+"/job."+job_id+".proxy";
    remove(fname.c_str());
    fname=user->ControlDir()+"/job."+job_id+".description";
    remove(fname.c_str());
    delete_job_id(); 
    return 1;
  };
  // call grid bank plugin/exec for consulting :)
  int bank_result;
  string bank_response;
  if(!(cont_plugins->run(job,*user,bank_result,bank_response))) {
    olog<<"Failed to run external plugin"<<endl;
    remove(rsl_fname.c_str());
    string fname=user->ControlDir()+"/job."+job_id+".proxy";
    remove(fname.c_str());
    fname=user->ControlDir()+"/job."+job_id+".state";
    remove(fname.c_str());
    fname=user->ControlDir()+"/job."+job_id+".description";
    remove(fname.c_str());
    delete_job_id();
    return 1;
  } else {
    if(bank_result != 0) {
      olog<<"External plugin returned non-zero result: "<<bank_result<<" "<<bank_response<<endl;
      remove(rsl_fname.c_str());
      string fname=user->ControlDir()+"/job."+job_id+".proxy";
      remove(fname.c_str());
      fname=user->ControlDir()+"/job."+job_id+".state";
      remove(fname.c_str());
      fname=user->ControlDir()+"/job."+job_id+".description";
      remove(fname.c_str());
      delete_job_id();
      return 1;
    };
  };
  if(bank_response.length()) olog<<"Plugin response: "<<bank_response<<endl;
  job_id.resize(0);
  return 0;
}

int JobPlugin::read(unsigned char *buf,unsigned long long int offset,unsigned long long int *size) {
  if(!initialized) return 1;
  return direct_fs->read(buf,offset,size);
}

int JobPlugin::write(unsigned char *buf,unsigned long long int offset,unsigned long long int size) {
  if(!initialized) return 1;
  if(!rsl_opened) { return direct_fs->write(buf,offset,size); };
  /* write to rsl */
  if(job_id.length() == 0) return 1;
  if(offset >= (sizeof(job_rsl)-1)) return 1;
  if(size >= (sizeof(job_rsl)-1)) return 1;
  if((offset+size) >= (sizeof(job_rsl)-1)) return 1;
  memcpy(job_rsl,buf+offset,size);
  return 0;
}

int JobPlugin::readdir(const char* name,list<DirEntry> &dir_list,DirEntry::object_info_level mode) {
  if(!initialized) return 1;
  if(name[0] == 0) { /* root jobs directory */
    DirEntry dent(false,"new");
    dir_list.push_back(dent);
    struct dirent file_;
    struct dirent *file;
    string cdir=user->ControlDir();
    DIR *dir=opendir(cdir.c_str());
    if(dir != NULL) for(;;) {
      readdir_r(dir,&file_,&file);
      if(file == NULL) break;
      int l=strlen(file->d_name);
      if(l>15) {
        if(!strncmp(file->d_name,"job.",4)) {
          if(!strncmp((file->d_name)+(l-6),".local",6)) { /* can only process those with .local information */
            JobLocalDescription job_desc;
            string fname=cdir+'/'+file->d_name;
            if(job_local_read_file(fname,job_desc)) {
              if(job_desc.DN == subject) {
                JobId id((file->d_name)+4,l-6-4);
                dir_list.push_back(DirEntry(false,id.c_str()));
              };
            };
          };
        };
      };
    };
    closedir(dir);
    return 0;
  };
  if(!strcmp(name,"new")) { /* directory /new is empty */
    return 0;
  };
  /* check for allowed job directory */
  const char* logname;
  string id;
  string log;
  if(is_allowed(name,false,&id,&logname,&log)) {
    if(logname) {
      if((*logname) != 0) {
        if(strchr(logname,'/') != NULL) return 1; /* no subdirs */
        id=user->ControlDir()+"/job."+id+"."+logname;
        struct stat64 st;
        if(::stat64(id.c_str(),&st) != 0) return 1;
        if(!S_ISREG(st.st_mode)) return 1;
        DirEntry dent(true,logname);
        dent.may_read=true;
        dir_list.push_back(dent);
        return -1;
      };
      DIR* d=::opendir(user->ControlDir().c_str());
      if(d == NULL) { return 1; }; /* maybe return ? */
      struct dirent *de;
      struct dirent de_;
      id="job."+id+".";
      int idl = id.length();
      for(;;) {
        readdir_r(d,&de_,&de);
        if(de == NULL) break;
        if((!strcmp(de->d_name,".")) || (!strcmp(de->d_name,".."))) continue;
        if(strncmp(id.c_str(),de->d_name,idl) != 0) continue;
        DirEntry dent(true,(de->d_name)+idl);
        dent.may_read=true;
        dir_list.push_back(dent);
      };
      ::closedir(d);
      return 0;
    };
    /* allowed - pass to file system */
    if(log.length() > 0) {
      char* s = strchr(name,'/');
      if((s == NULL) || (s[1] == 0)) {
        DirEntry dent(false,log.c_str());
        dent.may_dirlist=true;
        dir_list.push_back(dent);
      };
    };
    return direct_fs->readdir(name,dir_list,mode);
  };
  return 1;
}

int JobPlugin::checkdir(string &dirname) {
  /* chdir to /new will create new job */
  if(dirname.length() == 0) return 0; /* root */
  if(dirname == "new") { /* new job */
    if(!initialized) return 1;
    if(!make_job_id()) return 1;
    dirname=job_id;
    return 0;
  };
  const char* logname;
  string id;
  if(is_allowed(dirname.c_str(),false,&id,&logname)) {
    if(logname) {
      if((*logname) != 0) return 1; /* log directory has no subdirs */
      return 0;
    };
    if((dirname == id) && (proxy_fname.length())) {  /* cd to session directory - renew proxy request */
      string old_proxy_fname=user->ControlDir()+"/job."+id+".proxy";
      CertInfo new_ci(proxy_fname.c_str());
      time_t new_proxy_expires=new_ci.TimeLeft()+time(NULL);
      CertInfo old_ci(old_proxy_fname.c_str());
      time_t old_proxy_expires;
      if(!old_ci) { old_proxy_expires=time(NULL); }
      else { old_proxy_expires=old_ci.TimeLeft()+time(NULL); }
      if(new_proxy_expires > old_proxy_expires) {
        if(renew_proxy(old_proxy_fname.c_str(),proxy_fname.c_str()) == 0) {
          olog<<"New proxy expires at "<<mds_time(new_proxy_expires)<<endl;
          if(!job_restart_mark_put(JobDescription(id,""),*user)) {
            olog<<"Failed to report renewed proxy to job"<<endl;
          };
        } else {
          olog<<"Failed to renew proxy"<<endl;
        };
      };
    };
    return direct_fs->checkdir(dirname);
  };
  return 1;
}

int JobPlugin::checkfile(string &name,DirEntry &info,DirEntry::object_info_level mode) {
  if(!initialized) return 1;
  if(name.length() == 0) {
    info.name=""; info.is_file=false;
    return 0; 
  };
  if(!strcmp(name.c_str(),"new")) {
    info.name=""; info.is_file=false;
    return 0; 
  };
  const char* logname;
  string id;
  if(is_allowed(name.c_str(),false,&id,&logname)) {
    if(logname) {
      if((*logname) == 0) { /* directory itself */
        info.is_file=false; info.name=""; info.may_dirlist=true;
      }
      else {
        id=user->ControlDir()+"/job."+id+"."+logname;
        struct stat64 st;
        if(::stat64(id.c_str(),&st) != 0) return 1;
        if(!S_ISREG(st.st_mode)) return 1;
        info.is_file=true; info.name=""; info.may_read=true;
      };
      return 0;
    };
    return direct_fs->checkfile(name,info,mode);
  };
  return 1;
}

bool JobPlugin::delete_job_id(void) {
  if(job_id.length() != 0) {
    string dir=user->SessionRoot()+"/"+job_id;
    list<FileData> job;
    if(delete_all_files(dir,job,true) != 0) return false;
    remove(dir.c_str());
    job_id="";
  };
  return true;
}

bool JobPlugin::make_job_id(const string &id) {
  // claim id by creating session directory
  if(readonly) {
    olog<<"Not accepting new jobs"<<endl;
    return false;
  };
  string dir=user->SessionRoot()+"/"+id;
  if(mkdir(dir.c_str(),0700) != 0) return false;
  chmod(dir.c_str(),0700);
  fix_file_owner(dir,*user);
  // reject reserved words
  if((id == "new") || (id == "info")) return false;
  // check if job.id.status does not exist
  string fname=user->ControlDir()+"/job."+job_id+".status";
  struct stat st;
  if(stat(fname.c_str(),&st) == 0) {
    rmdir(dir.c_str());
    return false;
  };
  job_id=id;
  return true;
}

bool JobPlugin::make_job_id(void) {
  if(readonly) {
    olog<<"Not accepting new jobs"<<endl;
    return false;
  };
  int i;
  for(i=0;i<100;i++) {
    job_id=inttostring(rand(),1)+inttostring(rand(),1);
    string dir=user->SessionRoot()+"/"+job_id;
    if(mkdir(dir.c_str(),0700) != 0) continue;
    fix_file_owner(dir,*user);
    break;
  };
  if(i>=100) { job_id=""; return false; };
  return true;
}

/*
  name - name of file to access
  locked - return false if job already running
  jobid - returns id extracted from name
  logname - 
  log - stdlog of job
*/
bool JobPlugin::is_allowed(const char* name,bool locked,string* jobid,char const ** logname,string* log) {
  if(logname) (*logname) = NULL;
  JobId id(name);
  int n=id.find('/'); if(n != string::npos) id.erase(n);
  if(jobid) (*jobid)=id;
  JobLocalDescription job_desc;
  if(job_local_read_file(id,*user,job_desc)) {
    if(job_desc.DN == subject) {
      if(log) (*log)=job_desc.stdlog;
      if(logname && (n != string::npos)) {
        int l = job_desc.stdlog.length();
        if(l != 0) {
          if(strncmp(name+n+1,job_desc.stdlog.c_str(),l) == 0) {
            if(name[n+1+l] == 0) { (*logname)=name+n+1+l; return true; };
            if(name[n+1+l] == '/') { (*logname)=name+n+1+l+1; return true; };
          };
        };
      };
      if(!locked) return true;
      job_state_t status=job_state_read_file(id,*user);
      if((status == JOB_STATE_ACCEPTED) || (status == JOB_STATE_PREPARING)) {
        return true; 
      };
    };
  };
  return false;
}

